"""
Tests whether the save_to_file appears to be working,
bare minimum smoke test
"""


import os

from ansifier.ansifier import ImageFilePrinter


TEST_IMAGE_PATH = 'images-examples/catClout.png'
TEST_OUTPUT_TXT_PATH = 'tests/temp.txt'


def test_saves_to_file():
    """ 
    Checks that:
    * file was created
    * data was written to file
    DOES NOT check file contents for correctness otherwise
    """
    printer = ImageFilePrinter(TEST_IMAGE_PATH)

    printer.save_text(TEST_OUTPUT_TXT_PATH)

    assert os.path.exists(TEST_OUTPUT_TXT_PATH), 'expecting output file to have been created'
    assert os.path.getsize(TEST_OUTPUT_TXT_PATH) > 0, 'expecting output file to contain data'

    os.remove(TEST_OUTPUT_TXT_PATH)




"""
Main purpose of this file is to expose class ImageFilePrinter for consumption

TODO/ideas for improvements not yet made into GitHub issues:
    * ENHANCEMENT: Minimize output chars; only change color when you need to
    * enhancement: add a docstring here
    * enhancement: generalize a wrapper for _validate functions to dry code out
    * enhancement: more robust tests that valid inputs work
    * enhancement: test for expected exceptions with bad inputs!
    * enhancement: add utf-8 check to ansifier, warn if warranted
    * feature/enhancement: option: disable validation (performance optimization)
    * feature: save and load gif frames to and from files (proprietary format?)
    * feature: allow ImageFilePrinter to reload image file
    * feature: add ability to mandate output dimensions
       (scale to different aspect ratios)
"""


import html
import logging
from logging.handlers import RotatingFileHandler
import re
import signal

from PIL import Image
from colorama import just_fix_windows_console
from os import path, access, R_OK, W_OK
from shutil import get_terminal_size
from sys import exit
from time import sleep


from ansifier.config import CHARS, LOG_FILENAME, RESIZE_OPTIONS, LOG_LEVEL


# A couple of useful functions for hacking up ImageFilePrinter.output
#  post-generation

def remove_ansi_escape_sequences(text):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)

def number_of_spans_per_line(text):
    span_pattern = re.compile(r'<span\b[^>]*>(.*?)<\/span>')
    br = '<br/>'
    topline = text.split(br)[0]
    spanmatches = span_pattern.findall(topline)
    return len(spanmatches)

def length_after_processing(text, output_format='ansi-escaped'):
    if 'ansi' in output_format:
        processed_text = remove_ansi_escape_sequences(text).split('\n')[0]
        ret = len(processed_text)
    elif 'html' in output_format:
        ret = number_of_spans_per_line(text)
    return ret


# classes, main class for export is ImageFilePrinter,
# inherits from AnsifierBase,
# uses Cells to convert image pixels to output

class AnsifierBase():
    """
    contains input validation functions
    shared between Cell and ImageFilePrinter classes.
    Only intended to be subclassed.
    """


    def _validate_char_list(self, chars):
        """
        Checks:
        1. chars is a list
            may raise TypeError
        2. len(chars) >=1
            may raise ValueError
        """
        error_t = None
        error_message = None

        if not isinstance(chars, list):
            error_t = TypeError
            error_message = f'chars parameter must be a list, but {__class__} '\
                f'received {type(chars)}'

        elif not len(chars) >= 1:
            error_t = ValueError
            error_message = f'char list must have 1 or more element(s), '\
                f'but {__class__} received {chars}'

        if error_t is not None:
            self.logger.error(error_message)
            raise TypeError(error_message)


    def _validate_int(self, i_to_check, minimum=0, message=None):
        """
        Checks
        1. i_to_check can be cast to integer
        2. i_to_check is > minimum
        May raise a ValueError if either check fails
        """
        if message is None:
            message = f'param must be an integer > {minimum}, but {self.__class__} '\
                f'received {i_to_check}'
        try:
            i_to_check = int(i_to_check)
        except ValueError as e:
            self.logger.error(str(e) + message)
            raise e

        if not i_to_check > minimum:
            raise ValueError(message)

    
    def _validate_is_boolean(self, var_to_check, message=None):
        if message is None:
            message = 'parameter must be a boolean, '\
                f'but {self.__class__} received {type(var_to_check)}'
        if not isinstance(var_to_check, bool):
            self.logger.error(message)
            raise TypeError(message)


class Cell(AnsifierBase):
    """
    The intended public interface for this class is:
    * instantiation with at least a tuple of (r, g, b, a) integers,
      each integer being between 0 and 255 inclusive.
      Cell((r,g,b,a)) -> <Cell object>
    * After instantiation, conversion to str is the main point of this class:
      str(cell) -> "@@" (but with an ansi escape for color)
    * inheritance, I guess??

    One block of output, 2 characters wide and 1 tall, representing one pixel
    in the input image (after any scaling).
    Character is determined by either alpha (transparency, default)
    or by brightness (total of RGB values).
    Currently, each character is composed of
        a foreground color escape,
        2x the character itself
    resets (\\033[38;2;255;255;255m) are to be handled by the code
    that uses the Cell
    """

    def __init__(self, pixel, from_brightness=False, chars=None, output_format='ansi-escaped'):
        """
        :param pixel: tuple, of integers, length 4
            (red, green, blue, alpha) levels between 0 and 255 inclusive.
        :param from_brightness: bool, whether to use brightness to determine
            output glyph (defaults to False, which will use alpha)
        :param chars: list of 1-length strings, any non-zero length
            characters that will be selected from, intended to be ordered 
            from least to most visible when printed to a terminal
        :param output_format: str, how to convert pixel to char
            see Cell.output_formats
        """
        self.output_formats = {
                'ansi-escaped': self._ansi_escaped_output,
                'html/css': self._html_css_output
        }

        if chars is None:
            chars = CHARS
        #self._validate_char_list(chars)  # TODO how does this affect performance?
        self.chars = chars
        self.intervals = [255/len(self.chars)*i
                          for i in range(len(self.chars)-1, -1, -1)]
        # (how good is the interpreter at optimizing this? Does it??)

        self._validate_pixel(pixel)
        self.pixel = pixel

        self.output_format = output_format
        self.from_brightness = from_brightness
        self.r = self.pixel[0]
        self.g = self.pixel[1]
        self.b = self.pixel[2]
        self.a = self.pixel[3]
        self.char = self._get_char()


    def __str__(self):
        # assume a terminal character is about twice as tall as it is wide
        output_function = self.output_formats[self.output_format]
        return output_function()

    def _ansi_escaped_output(self):
        color_escape = f"\033[38;2;{self.r};{self.g};{self.b}m" 
        if self.char == ' ':  # no need to color-escape transparent cells
            ret = self.char * 2
        else:
            ret = color_escape + self.char * 2
        return ret

    def _html_css_output(self):
        html_escaped_char = html.escape(self.char)
        if self.char == ' ':  # no need to color transparent cells
            ret = '&nbsp;' * 2  # html.escape does not handle spaces...
            # may run into more trouble with other characters
        # may need to deal with other escapes
        else:
            color_char = f'<span style="color: rgba({self.r},{self.g},{self.b},{self.a})">{html_escaped_char}</span>' 
            ret = color_char * 2
        return ret


    def _validate_pixel(self, pixel):
        """
        Checks:
        1. pixel is a tuple
            may raise TypeError
        2. pixel has 4 members
            may raise ValueError
        3. each member of pixel is an integer between 0 and 255
            may raise ValueError
        """
        error_t = None
        error_message = None
        if not isinstance(pixel, tuple):
            error_t = TypeError
            error_message = f'pixel parameter must be a tuple, but '\
                f'{__class__} received a {type(pixel)}'
        elif not len(pixel) == 4:
            error_t = ValueError
            error_message = f'pixel parameter must have 4 members, but '\
                f'{__class__} received {len(pixel)}'
        elif not all(isinstance(v, int) for v in pixel) \
         and not all(map(lambda el: 0 <= el and el <= 255, pixel)):
            error_t = ValueError
            error_message = f'pixel parameter elements must each be an '\
                f'integer >=0 and <=255, but {__class__} received {pixel}'

        if error_t is not None:
            self.logger.error(error_message)
            raise TypeError(error_message)


    def _validate_output_format(self, output_format):
        error_t = None
        error_message = None
        if not output_format in self.output_formats:
            error_t = ValueError
            error_message = f'{output_format} is not an implemented output format, '\
                f'try one of these: {list(self.output_formats.keys())}'
        if error_t is not None:
            self.logger.error(error_message)
            raise error_t(error_message)



    def _get_char(self):
        if self.from_brightness:
            return self._get_char_from_brightness()
        else:
            return self._get_char_from_alpha()


    def _get_char_from_brightness(self):
        """ use sum of RGB values to determine char """
        char = None
        brightness = self.pixel[0] + self.pixel[1] + self.pixel[2]  # max. 765
        for i, interval in enumerate(self.intervals):
            if brightness >= interval*3:
                char = self.chars[i]
                break

        return char


    def _get_char_from_alpha(self):
        """ use transparency to determine char """
        char = None
        for i, interval in enumerate(self.intervals):
            if self.a >= interval:
                char = self.chars[i]
                break

        if self.a < 20:  # prevent weird artifacting from resize...
            char = ' '

        return char


class ImageFilePrinter(AnsifierBase):
    """
    The intended public interface for this class is
    * Instantiation with at least an image path, the image to turn into text
      ImageFilePrinter('/path/to/image.png') -> <ImageFilePrinter>.
      Images are loaded into memory and converted to text output
      on object initialization.
    * Calling save_text(path) will
      dump text output to a file at path.
    * Calling print_text() will
      dump text output to stdout.
    * Calling unload_image() will unload the Image object from memory,
      but will NOT unload stored text output.
      Default behavior is to unload image after output was generated in __init__
    * Calling load_image() will load the Image object back into memory
      from the file at self.image_path
    * Calling generate_output() will always populate self.output if an image
      is loaded, and will also populate self.frames when self.animate is truthy.
      If generate_output() is called with no image loaded, nothing happens.
      

    Manipulating attributes in any other way is unsupported, at least for now.
    Dynamic changes to max_height and max_width, etc., are likely to be
    supported in the future.

    a couple of important pieces of state:
    * self.output holds the string to dump to stdout or file for static images.
      It is built when an image is loaded, on initialization or otherwise.
      ! The exception here is that animations will update this attribute as
        the object converts frames into strings to dump. Only the most recently
        converted frame is stored in this attr (NOT the most recently dumped).
    * self.frames is like self.output except it stores a list, with one element
      for each ansified frame in the last animated image that was loaded.
    """

    reset_escape = "\033[38;2;255;255;255m"
    html_opening_tags = '<div style="font-family: monospace; line-height: 1.2;">'
    html_closing_tags = '</div>'

    def __eq__(self, other) :
        return self.__dict__ == other.__dict__

    def __init__(self, image_path, max_height=None, max_width=None,
            resize_method=Image.LANCZOS, char_by_brightness=False, 
            chars=CHARS, animate=0, loop_infinitely=False, output_format="ansi-escaped",
            logfile=None):
        """
        :param image_path: str, path to image to represent as text.
            Supported image formats are those supported by PIL version 9
        :param max_height: int, restrict output text to this number of rows
            defaults to terminal rows - 1 (to account for most prompts)
        :param max_width: int, restrict output text to this width in characters;
            One cell is two characters wide, so 1 is not a valid max_width.
            Defaults to terminal columns // 2
        :param resize_method: PIL.Image.<ResamplingMethod>, see PIL version 9;
            integer >= 0 and <= 5.
        :param char_by_brightness: bool, whether to use brightness to determine
            output glyphs (defaults to False, use alpha instead)
        :param chars: list of at least one string where each string has len 1.
            Characters that will be selected from, intended to be ordered 
            from least to most visible when printed to a terminal
        :param animate: int, truthiness determines whether to read keyframes
            from animated .gif image inputs and print each keyframe. If truthy,
            the int value is interpreted as the number of milliseconds to sleep
            between dumping each frame to stdout, not accounting for the time
            it takes to perform the output i.e. actual delay between frames
            may be longer depending on output dimensions, hardware, terminal
            emulator, etc.
            It may be advisable to set this value to a factor of your monitor's
            refresh rate to avoid your monitor catching ansifier mid-print.
        :param loop_infinitely: bool, whether to wait for keyboardinterrupt
            to stop animating
        :param logfile: str, where to store log file
        :param output_format: str, defaults to "ansi-escaped", see ImageFilePrinter.output_formats
        """
        self.logger = None
        if logfile is None:
            logfile = path.join(path.dirname(__file__), '..', LOG_FILENAME)
        self._validate_can_write_file(logfile, log=False)
        self.logger = self._initialize_logger(logfile)

        self.image_object = None
        self.output = None
        self.output_width = None
        self.output_height = None
        self.frames = []

        type_check_message = '{} must be a {}, '\
            + f'but {self.__class__} received a' + '{}'

        self._validate_char_list(chars)
        self.chars = chars
        self.intervals = [255/len(CHARS)*i for i in range(len(CHARS)-1, -1, -1)]

        # validated in load_image
        self.image_path = image_path

        valid_resize_methods = [val[0] for val in RESIZE_OPTIONS.values()]
        valid_resize_names = [val[1] for val in RESIZE_OPTIONS.values()]
        if resize_method not in valid_resize_methods:
            message = f'Invalid resize method: {resize_method}\n'\
                f'options are {zip(valid_resize_methods, valid_resize_names)}'
            self.logger.error(message)
            raise ValueError(message)
        self.resize_method = resize_method

        if animate != 0:
            self._validate_int(animate,
                message=type_check_message.format(
                    'animate', 'non-negative integer', type(animate)))
        self.animate = animate

        self._validate_is_boolean(char_by_brightness,
            type_check_message.format(
                'char_by_brightness', 'boolean', type(char_by_brightness)))
        self.char_by_brightness = char_by_brightness

        self._validate_is_boolean(loop_infinitely,
            type_check_message.format(
                'loop_infinitely', 'boolean', type(loop_infinitely)))
        self.loop_infinitely = loop_infinitely

        self.logger.info(f'reading image file {image_path}')
        self.logger.debug(f'working with these chars: {self.chars} '
            '(len {len(self.chars)})')
        self.logger.debug(f'working with these intervals: {self.intervals} '
            '(len {len(self.intervals)})')

        Cell((0,0,0,0))._validate_output_format(output_format)
        self.output_format = output_format

        if max_height is None or max_width is None:
            self.logger.debug('Getting terminal dimensions')
        if max_height is None:
            max_height = get_terminal_size().lines - 1  # -1 for prompt
        if max_width is None:
            # assume a terminal character is about twice as tall as it is wide
            max_width = get_terminal_size().columns
        self._validate_dimensions(max_height, max_width)
        self.max_height = max_height
        self.max_width = max_width // 2

        self.load_image()  # loads image into memory
        self.generate_output()  # converts self.image_object to string output
        self.logger.info('generated output with '
            f'w:{self.output_width}, h:{self.output_height}')
        self.unload_image()


    def print_text(self):
        """
        If self.animate is set,
        pass ALL responsibility to self._print_animated.
        Else, 
        Dump static output to stdout.
        """
        # checks for windows internally
        just_fix_windows_console()

        if self.animate:
            self._print_animated()

        else:
            rows = self.output.count("\n")
            self.logger.info(f'dumping {len(self.output)} chars in {rows} rows derived from '
                f'{self.image_path} to stdout')
            #sys.stdout.write(self.output)
            #sys.stdout.flush()  # TODO this might be marginally faster
            print(self.output, end='')


    def _print_animated(self):
        """
        Print each frame to stdout in a loop,
        forever if self.loop_infinitely... TODO a duration might be a nice option?
        """
        self._register_sigint_handler()
        done = False
        frame_interval = self.animate / 1000.0
        self.logger.info(f'dumping {len(self.frames)} frames from '
        f'{self.image_path} to stdout with {frame_interval}s between frames')
        while not done:
            # important to minimize overhead in this for loop...
            # maybe a good idea to write this in C?
            for text_frame in self.frames:
                #self.logger.debug(f'dumping {len(text_frame)} chars derived '
                    #f'from {self.image_path} to stdout')
                #sys.stdout.write(text_frame)
                #sys.stdout.flush()  # TODO this might be marginally faster
                print(text_frame, end='')
                sleep(frame_interval)
            if not self.loop_infinitely:
                done = True


    def save_text(self, filepath):
        """
        Validates filepath, then,
        if self.animate is set,
        pass ALL responsibility to self._save_animated,
        which is not yet implemented... so does nothing.
        Else, 
        Dump static output to filepath.
        """
        self._validate_can_write_file(filepath)
        if self.animate:
            #self.save_animated()  # TODO dump and load gif frames!
            self.logger.error(
                f'Saving animated GIF frames to file not yet supported')
        else:
            self.logger.info(f'dumping {len(self.output)} chars derived from '
                f'{self.image_path} to {filepath}')
            self._save_file(filepath)


    def _save_file(self, filepath):
        """
        dumps self.output to filepath.
        filepath assumed to have been validated in init,
        self.output assumed to have been generated before call
        """
        with open(filepath, 'w') as wf:
            for c in self.output:
                wf.write(c)


    def load_image(self):
        self._validate_can_read_file(self.image_path)
        try:
            # TODO handle DecompressionBombWarning for large images
            # this holds the file descriptor open for the lifetime of
            # the image object afaik, which seems necessary for seeks.
            self.image_object = Image.open(self.image_path, 'r')
        except Exception as e:
            self.logger.critical(e)
            raise e


    def generate_output(self):
        if self.image_object is None:
            self.logger.warning('generate_output() was called '
                'with no image loaded')
            return
        if self.animate:
            self.frames = self._convert_animation_to_frames()
        else:
            self.output = self._convert_image_to_string(
                self._prepare_for_dump(self.image_object))
        self.output_width = length_after_processing(
            self.output, self.output_format)
        self.output_height = self.output.count('\n')


    def unload_image(self):
        """
        dellocates image object, but not the text output it generated
        """
        if self.image_object is not None:
            self.image_object.close()
        else:
            self.logger.warning('unload_image() called with no image loaded')
        self.image_object = None


    def _register_sigint_handler(self):
        """
        prints an ansi reset before doing the usual thing
        """
        def handler(signum, frame):
            print(ImageFilePrinter.reset_escape, end='')
            signal.default_int_handler(signum, frame)
        signal.signal(signal.SIGINT, handler)


    def _convert_animation_to_frames(self):
        self.logger.info(f'Processing frames from {self.image_path}...')
        ret = []

        while True:
            self.logger.debug(f'Processing frame {self.image_object.tell()}')
            # must be done per-seek
            image_frame = self._prepare_for_dump(self.image_object)
            # as a side effect, this will update self.output, but
            # I don't think this matters much
            ret.append(self._convert_image_to_string(image_frame))

            try:
                self.image_object.seek(self.image_object.tell() + 1)
            except EOFError:  # TODO use n frames from image_object
                self.logger.debug(f'No more frames to process from image')
                self.image_object.seek(0)
                break

        return ret


    def _prepare_for_dump(self, image):
        """
        converts a PIL.Image to RGBA and resizes it according to
        self.max_width and self.max_height.
        Returns the modified Image object.
        """
        image = image.convert('RGBA')
        self.logger.debug(f'before: {image.size}')
        image.thumbnail((self.max_width, self.max_height),
                self.resize_method)
        self.logger.debug(f'target: {(self.max_width, self.max_height)}')
        self.logger.debug(f'after: {image.size}')
        return image


    def _convert_image_to_string(self, image):
        """
        Intent is that this should be the ONLY place self.output is assigned to.
        Call this when you need to dump output chars, but self.output is None;
        Only convert Image objects (incl. frames) to strings once if possible
        for the lifetime of the ImageFilePrinter.
        Also returns the output string in order to allow frame caching for
        animations.
        """
        self.logger.debug(f'converting {image} to ansi-escaped string')
        output = ''
        for j in range(image.size[1]):
            for i in range(image.size[0]):
                pix = image.getpixel((i, j))
                cell = Cell(pix, self.char_by_brightness, self.chars, self.output_format)
                output += str(cell)
            output = self._finish_line(output)
        output = self._finish_output(output)

        self.output = output
        self.logger.debug(f'done converting {image}')
        return output


    def _finish_line(self, output):
        if 'ansi' in self.output_format:  # horrible hack, forgive me
            output += '\n'
        elif 'html' in self.output_format:
            output += '<br/>'
        return output


    def _finish_output(self, output):
        if 'ansi' in self.output_format:  # horrible hack, forgive me
            output += ImageFilePrinter.reset_escape
        elif 'html' in self.output_format:
            output = ImageFilePrinter.html_opening_tags\
                   + output\
                   + ImageFilePrinter.html_closing_tags
        return output


    def _validate_can_read_file(self, filepath):
        """
        checks three conditions:
        1. filepath exists
            may raise a FileNotFoundError
        2. filepath is not a directory (this should mean it's a file right?)
            may raise a IsADirectoryError
        3. filepath is readable (TODO this is redundant right?)
            may raise a PermissionError
        """
        error_t = None
        error_message = None
        try:
            filepath = path.abspath(filepath)
        except TypeError as e:
            self.logger.error(e)
            raise e

        if not (path.exists(filepath)):
            error_t = FileNotFoundError
            error_message = f'Unable to read file {filepath}: does not exist '\
                '(or is behind a privileged directory)'
        elif path.isdir(filepath):
            error_t = IsADirectoryError
            error_message = f'Unable to read file {filepath}: '\
                f'that\'s a directory, not a file'
        elif not access(filepath, R_OK):
            error_t = PermissionError
            error_message = f'Unable to read file {filepath}: '\
                'insufficient permissions'

        if error_t is not None:
            self.logger.error(error_message)
            raise error_t(error_message)


    def _validate_dimensions(self, height, width):
        """
        For each dimension, checks:
        1. dimension is an integer
        2. dimension is > 0
        May raise a ValueError if either check fails
        """
        self._validate_int(height,
            message='requested height must be an integer > 0, '
                    f'but {__class__} received {height}')
        self._validate_int(width, 1,
            message='requested width must be an integer > 1, '
                    f'but {__class__} received {width}')


    def _validate_can_write_file(self, filepath, log=True):
        """
        checks three conditions:
        1. filepath's parent directory exists
            may raise FileNotFoundError
        2. filepath is not a directory (this should mean it's a file right?)
            may raise IsADirectoryError
        3. filepath's parent directory is writable
            may raise PermissionError
        """
        error_t = None
        error_message = None
        try:
            filepath = path.abspath(filepath)
        except TypeError as e:
            if log:
                self.logger.error(e)
            raise e
        parent_dir = path.dirname(filepath)

        if not (path.exists(parent_dir)):
            error_t = FileNotFoundError
            error_message = f'Unable to write to {filepath}: {parent_dir} '\
                'does not exist (or is behind a privileged directory)'
        elif path.isdir(filepath):
            error_t = IsADirectoryError
            error_message = f'Unable to write to file {filepath}: '\
                f'that\'s a directory, not a file'
        elif not access(parent_dir, W_OK):
            error_t = PermissionError
            error_message = f'Unable to write to file {filepath}: '\
                'insufficient permissions'

        if error_t is not None:
            if log:
                self.logger.error(error_message)
            raise error_t(error_message)



    @staticmethod
    def _initialize_logger(logpath):
        """
        By default,
        logs to file at logpath with severity INFO or greater,
        logs to stdout with severity ERROR or greater.
        logpath was validated in init.
        """
        logger = logging.getLogger(__name__)
        logger.setLevel(LOG_LEVEL)

        log_formatter = logging.Formatter(
            '%(asctime)s;%(levelname)s;%(message)s')

        log_file_handler = RotatingFileHandler(
            filename=logpath,
            maxBytes=10000000,  # 10 MB
            backupCount=2)
        log_file_handler.setFormatter(log_formatter)

        log_stdout_handler = logging.StreamHandler()
        log_stdout_handler.setLevel(logging.ERROR)
        log_stdout_handler.setFormatter(log_formatter)

        logger.addHandler(log_file_handler)
        logger.addHandler(log_stdout_handler)

        return logger




"""
Tests whether writing to standard output is working,
bare minimum smoke test
"""


from io import StringIO
from contextlib import redirect_stdout

from ansifier.ansifier import ImageFilePrinter


TEST_IMAGE_PATH = 'images-examples/catClout.png'


def test_prints_to_stdout():
    """ 
    Checks that something was written to stdout
    DOES NOT check output for correctness otherwise
    """
    captured_output = StringIO()
    printer = ImageFilePrinter(TEST_IMAGE_PATH)

    with redirect_stdout(captured_output):
        printer.print_text()

    assert captured_output.getvalue()




"""
Tests whether ImageFilePrinter produces output within the requested dimensions
for images with landscape, portrait, and square aspect ratios.
Could use inheritance for test classes but would probably be overwrought
"""


import os
from unittest import TestCase

from ansifier.ansifier import ImageFilePrinter


class TestSquare(TestCase):
    image_path = 'images-examples/catSwag.png'

    def test_width_square(self):
        """ 
        Checks that output is exactly the requested maximum width for images
        with square aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_width=20)
        assert printer.output_width == 20

    def test_height_square(self):
        """ 
        Checks that output is exactly the requested maximum height for images
        with square aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_height=20)
        assert printer.output_height == 20


class TestWide(TestCase):
    image_path = 'images-examples/loaf.gif'

    def test_width_wide(self):
        """ 
        Checks that output is within the requested maximum width for images
        with wide aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_width=20)
        assert printer.output_width <= 20

    def test_height_wide(self):
        """ 
        Checks that output is within the requested maximum height for images
        with wide aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_height=20)
        assert printer.output_height <= 20


class TestTall(TestCase):
    image_path = 'images-examples/miku.gif'

    def test_width_tall(self):
        """ 
        Checks that output is within the requested maximum width for images
        with tall aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_width=20)
        assert printer.output_width <= 20

    def test_height_tall(self):
        """ 
        Checks that output is within the requested maximum height for images
        with tall aspect ratios
        """
        printer = ImageFilePrinter(self.image_path, max_height=20)
        assert printer.output_height <= 20




import sys
import os
import pytest


REPO_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


@pytest.fixture(scope="session", autouse=True)
def add_repo_root_to_sys_path():
    """
    Ensure that the repository root is in PYTHONPATH for the entire test session.
    """
    sys.path.insert(0, REPO_ROOT)

    yield

    # Optionally, clean up by removing the repo_root from sys.path after tests
    sys.path.remove(REPO_ROOT)



"""
Tests whether html output is working as expected
in the most basic case
kind of a smoke test
"""

from io import StringIO
from contextlib import redirect_stdout
from difflib import SequenceMatcher

from ansifier.ansifier import ImageFilePrinter


TEST_IMAGE_PATH = 'images-examples/catClout.png'
expected_output_file = 'tests/expected.html'
with open(expected_output_file, 'r') as rf:
    expected_output = rf.read()


def test_html_output():
    """ 
    Checks that something was written to stdout
    DOES NOT check output for correctness otherwise
    """
    captured_output = StringIO()
    printer = ImageFilePrinter(TEST_IMAGE_PATH, output_format='html/css', max_height=20)

    with redirect_stdout(captured_output):
        printer.print_text()
    observed_output = captured_output.getvalue()

    print(f'output matches {SequenceMatcher(a=expected_output, b=observed_output).ratio() * 100}%')
    assert observed_output == expected_output




#!/bin/python


"""
This file implements the CLI to ansifier, as well as a handful of utility
functions which make sense in the context of a terminal -
I can see these functions being used in other contexts, however,
so they might become ImageFilePrinter member functions sooner or later.
"""


import argparse
import os
from importlib.metadata import version
from shutil import get_terminal_size
from subprocess import check_output
from sys import version_info

from ansifier.ansifier import ImageFilePrinter, length_after_processing
from ansifier.ansifier import Cell
from ansifier.config import CHARS, RESIZE_OPTIONS

output_formats = [format for format in Cell((0,0,0,0)).output_formats.keys()]


def center_frame_horizontally(frame, width, line_len=None):
    """
    this is only intended for use in ansi-escaped output mode
    :param frame: str, image printer output
    :param width: int, num of columns in container (terminal)
    :param line_len: int, number of characters in a line of output
        post-escaping; pass this in if it's already been calculated,
        otherwise we'll calculate it for you
    :return: frame, modified to be centered based on width
    """
    line_len = length_after_processing(frame.split('\n')[0])
    frame = '\n'.join(
        [' ' * ((width - line_len) // 2) + l\
        for l in frame.split('\n')][:-1])
    frame += '\n'
    return frame


def center_frame_vertically(frame, height, line_len=None, num_lines=None):
    """
    this is only intended for use in ansi-escaped output mode
    :param frame: str, image printer output
    :param height: int, num of rows in container (terminal)
    :return: frame, modified to be centered based on height
    :param line_len: int, number of characters in a line of output
        post-escaping; pass this in if it's already been calculated,
        otherwise we'll calculate it for you
    :param num_lines: int, number of rows in output; pass this in if
        it's already been calculated, otherwise we'll calculate it for you
    """
    if line_len is None:
        line_len = length_after_processing(frame.split('\n')[0])
    if num_lines is None:
        num_lines = len(frame.split('\n'))
    difference = height - num_lines
    pad_num = difference // 2
    pad_line = ' ' * line_len + '\n'
    pad = pad_line*pad_num
    frame = pad_line + frame
    frame = pad + frame + pad
    return frame


def run_cli(args=None):
    """
    :param args: argparse.Namespace, arguments to use;
        should be generated by get_args(), which is what will happen
        if this argument is not specified
    :return: ImageFilePrinter, initialized using args
    Instantiates an ImageFilePrinter using argv
    May save ImageFilePrinter.output to a file if instructed
    DOES NOT print the text to the terminal;
    """
    global CHARS
    global RESIZE_OPTIONS
    NUM_CHARS = len(CHARS)

    if args is None:
        args = get_args()
    
    resize_method = None
    try:
        resize_method = RESIZE_OPTIONS[args.resize_method][0]
    except KeyError:
        print('error: bad value passed to resize_method switch; '
            f'see `{__file__} -h`.')
        exit(1)
    limit = args.limit_high_chars
    negative_limit = args.limit_low_chars
    #offset = args.char_offset

    if limit + negative_limit >= NUM_CHARS - 1:
        print(f'error: total of character limit arguments must be < '
            f'{NUM_CHARS-1} to get any visible output; see `{__file__} -h`.')
        exit(1)
    CHARS = CHARS[limit:]
    if negative_limit != 0:
        CHARS = CHARS[:-1*negative_limit - 1]
        CHARS.append(' ')
    if args.invert:
        CHARS = CHARS[-1::-1]
    #if args.no_whitespace:
        #CHARS.remove(' ')
    #INTERVALS = [255/len(CHARS)*i for i in range(len(CHARS)-1, -1, -1)]
    #if offset != 0:
        #INTERVALS = [i+INTERVALS[-1 - offset] for i in INTERVALS]
        #new_offset_addition = INTERVALS[-1*offset]
        #for i, interval in enumerate(INTERVALS):
            #print(f'offset of {offset} adds {new_offset_addition} to {interval}')
            #new_interval = interval+new_offset_addition
            #if new_interval <= 255:
                #INTERVALS[i] = new_interval
    #main(args.imageFilePath, args.max_height, args.max_width)

    image_printer = ImageFilePrinter(
        image_path=args.image_path,
        max_height=args.max_height,
        max_width=args.max_width,
        resize_method=resize_method,
        char_by_brightness=args.char_by_brightness,
        chars=CHARS,
        animate=args.animate,
        loop_infinitely=args.loop_infinitely,
        output_format=args.output_format)

    sz = get_terminal_size()
    width = sz.columns
    height = sz.lines

    line_len = length_after_processing(image_printer.output.split('\n')[0])
    if args.center_horizontally and width > line_len+1:
        if not args.meofetch:
            if args.animate:
                for i, frame in enumerate(image_printer.frames):
                    image_printer.frames[i] = center_frame_horizontally(
                        frame, width, line_len)
            else:
                image_printer.output = center_frame_horizontally(
                        image_printer.output, width, line_len)
        else:
            pass  # meofetch handles this

    num_lines = len(image_printer.output.split('\n'))
    if args.center_vertically and height > num_lines:
        if args.animate:
            for i, frame in enumerate(image_printer.frames):
                image_printer.frames[i] = center_frame_vertically(
                    frame, height, line_len, num_lines)
        else:
            image_printer.output = center_frame_vertically(
                image_printer.output, height, line_len, num_lines)
    
    if args.save_to_file:
        image_printer.save_text(args.save_to_file)

    return image_printer


def get_parser():
    """
    :return: argparse.ArgumentParser, the default ansifier argument parser
    """
    cell_info_message = 'note that a cell is roughly square, i.e. it is '\
        '2 terminal characters wide and 1 terminal character tall.'
    argparser = argparse.ArgumentParser(
        prog='ansifier',
        description='In its most basic usage, takes an image file as input and '
        'prints a unicode representation of the image to the terminal, using '
        'ansi escapes for color and determining what character to use based '
        'on the transparency of the region of the image represented by that '
        'character. By default, the image is scaled to the maximum dimensions '
        'that will fit within the terminal calling this program.')

    argparser.add_argument('-v', '--version', action='store_true',
        required=False, default=False, help='print version information and exit')

    argparser.add_argument('-H', '--max-height', action='store', type=int,
        required=False, default=None,
        help='Restrict output to this many rows at most; ' + cell_info_message)

    argparser.add_argument('-W', '--max-width', action='store', type=int,
        required=False, default=None,
        help='Restrict output to this many columns at most; ' + cell_info_message)

    argparser.add_argument('-c', '--limit-high-chars', action='store', type=int,
        required=False, default=0, help='remove <arg> chars from high (opaque) '
        f'output options; Available chars are: {CHARS}')

    argparser.add_argument('-C', '--limit-low-chars', action='store',
            type=int, required=False, default=0, help='remove <arg> chars '
                f'from low (transparent) output options; preserves space char. '
                f'Available chars are: {CHARS}')

    argparser.add_argument('-F', '--output-format', action='store',
            type=str, required=False, default='ansi-escaped', help='how to '
                'format output text - must be one of the following:\n'
                f'{output_formats}')

    # TODO this is a fun idea, but tricky in combination with other options
    #argparser.add_argument('-o', '--char-offset', action='store',
            #type=int, required=False, default=0, help='shift <arg> chars '
                #'higher in the array of available chars, without dropping the '
                #'space char (retains fully transparent cells). This is a WIP '
                #'and bad values are likely to break the program. '
                #f'Available chars are: {CHARS}')

    readable_resize_options = {(k, v[1]) for k, v in 
        RESIZE_OPTIONS.items()}
    argparser.add_argument('-r', '--resize-method', action='store', type=str,
        required=False, default='lz', help='algorithm used for resampling '
            'image to desired output dimensions. Defaults to "lz", Lanczos, '
            'which tends to work best when scaling images down to normal '
            'terminal dimensions. Options are the keys of this dict: '
            f'{readable_resize_options}')

    argparser.add_argument('-a', '--animate', action='store',
        required=False, type=int, default=0,
        help='If the input image is animated (.gif), process all keyframes and '
            'print them with ANIMATE milliseconds of delay between frames. '
            'This option is incompatible with -f. '
            'It is advisable to set this value to some factor of your monitor\'s '
            'refresh rate to avoid your monitor catching ansifier mid-print.')

    argparser.add_argument('-f', '--save-to-file', action='store', type=str,
        required=False, default=None, help='Write output to a file. '
            'Does not suppress terminal output. Will create or overwrite the '
            'file if needed, but will not create new directories.')

    argparser.add_argument('-L', '--loop-infinitely', action='store_true',
        required=False, default=False,
        help='With -a, causes the animation to loop until the '
        'program is terminated.')

    argparser.add_argument('-b', '--char-by-brightness', action='store_true',
        required=False, default=False, help='Use brightness (instead of '
            'alpha) to determine character used to represent an input region '
            'in output.')

    argparser.add_argument('-i', '--invert', action='store_true',
        required=False, default=False,
        help='Invert the effect of transparency (or brightness when using -b '
            '(--char-by-brightness) on char selection; useful for images with '
            'dark foregrounds and bright backgrounds, for example')

    argparser.add_argument('-z', '--center-horizontally', action='store_true',
        required=False, default=False,
        help='Use terminal size to center output horizontally. Only affects '
            'stdout, does not affect saved file contents if any')

    argparser.add_argument('-V', '--center-vertically', action='store_true',
        required=False, default=False,
        help='Use terminal size to center output vertically. Only affects '
            'stdout, does not affect saved file contents if any')

    argparser.add_argument('-m', '--meofetch', action='store_true',
            required=False, default=False,
            help='meofetch takes ansifier output and interleaves it '
                'with neofetch system info, with some padding.\n'
                'if the terminal isn\'t large enough, neofetch lines may be '
                'trimmed from the bottom, and/or the ansi image may not display.\n'
                'Note that the --animate option has no effect here. '
                'Note that using --center-horizontally in this context is '
                'discouraged.\n'
                'Finally, note that --max-width and --max-height may only reduce '
                'the output dimensions from those found based on terminal size.\n')

    # TODO not a bad idea but surprisingly not working?
    #argparser.add_argument('-w', '--no-whitespace', action='store_true',
        #required=False, default=False,
        #help='if this flag is set, disallow whitespace '
            #'(completely transparent blocks) in output.')

    argparser.add_argument('image_path', nargs='?', default=None)

    return argparser


def get_args(argparser=None):
    """
    :param argparser: argparse.ArgumentParset, should be retrieved from
        get_parser(), which is what will happen if this argument
        is not set
    :return: argparse.Namespace, result of argparser.parse_args()
    """
    if argparser is None:
        argparser = get_parser()
    args = argparser.parse_args()
    return args


def meofetch(args):
    """
    takes args object as returned by get_args()
    A bit of a scrappy, hacked up little script, at least for now.
    * Steals output from neofetch --off
    * generates ansi art from the image, takes terminal size into account
    * combines these outputs and prints it all

    TODO make this less of a horrible hack
    """
    args.animate = 0
    # TODO flesh these out and work them into argparse setup?
    LEFT_PADDING = -1  # -1 for center
    #LEFT_PADDING = 8  # more traditional
    BETWEEN_PADDING = 7

    # get terminal dimensions
    sz = get_terminal_size()
    w_orig = sz[0]
    h = sz[1] - 1  # -1 for prompt line

    # get output from neofetch
    system_info = check_output(['neofetch', '--off']).decode('UTF-8')
    sysinfo_lines = system_info.split('\n')
    sysinfo_lines = [' '*BETWEEN_PADDING + l for l in sysinfo_lines]

    # take lines off the bottom of neofetch's output until the output will fit in the terminal
    while h < len(sysinfo_lines):
        sysinfo_lines.pop()
    # TODO de-prioritize neofetch lines for screen space vs. ansi art here?
    # Account for horizonal room for neofetch output
    w = (w_orig - max(list(map(length_after_processing, sysinfo_lines)))) - 6

    if w > 0 and h > 0:

        if args.max_width is not None:
            args.max_width = min(args.max_width, w)
        else:
            args.max_width = w

        if args.max_height is not None:
            args.max_height = min(args.max_height, h) if args.max_width is not None else h
        else:
            args.max_height = h

        # the .output attr was not originally intended as a public interface,
        #  but it sure works as one
        ansi_art = run_cli(args).output
        ansi_lines = ansi_art.split('\n')[:-1]
        max_ansi_line_len = max(list(map(length_after_processing, ansi_lines)))
        if args.center_horizontally and w > max_ansi_line_len+1:
            ansi_lines = [' ' * ((w - max_ansi_line_len) // 2) + l for l in ansi_lines]
            ansi_art += '\n'
        # pad left
        #if LEFT_PADDING != -1:
            #ansi_lines = [' '*LEFT_PADDING + line for line in ansi_lines]
        #else:  # center ansi output horizontally
            # get the longest total line...
            #longest_output_line = max(map(length_after_processing,
                #[l1+l2 for l1, l2 in zip(ansi_lines, sysinfo_lines)]))
            #LEFT_PADDING = (w_orig - longest_output_line) // 2
            #ansi_lines = [' '*LEFT_PADDING + line for line in ansi_lines]  # TODO
    else:
        ansi_lines = []
        max_ansi_line_len = 0

    # ensure neofetch output does not unindent if it has more rows than ansi art
    # also centers ansi art in this case
    # TODO use a generator to alternate padding direction?
    i = 0
    padding_row = ' ' * (max_ansi_line_len + LEFT_PADDING)
    while len(sysinfo_lines) > len(ansi_lines):
        if i % 2:
            ansi_lines.insert(0, padding_row)
        else:
            ansi_lines.append(padding_row)
        i += 1

    # ensure neofetch output is centered when ansi art has more rows
    # TODO this is repeated in normal cli almost exactly
    difference = len(ansi_lines) - len(sysinfo_lines) + 2
    pad_num = difference // 2
    pad = [' ' for i in range(pad_num)]
    sysinfo_lines = pad + sysinfo_lines + pad

    for ansi_line, sysinfo_line in zip(ansi_lines, sysinfo_lines):
        # ensure no ansi lines are shorter
        #  seemingly due to how ansifier works... TODO?
        ansi_line = ansi_line.ljust(max_ansi_line_len)
        print(ansi_line, sysinfo_line)

    # This is necessary due to something about the output
    #  from neofetch... but why??? Something about cooked vs raw chars?
    print(os.popen('tput cnorm').read(), end='')
    print(os.popen('setterm --linewrap on').read(), end='')


def ansifier_cli_main():
    try:
        parser = get_parser()
        args = get_args(parser)
        if args.version:
            print('ansifier', version('ansifier'), 'from',
                os.path.dirname(__file__),
                f'(python {version_info[0]}.{version_info[1]})')
            exit(0)
        elif args.image_path is None:
            parser.error('the following arguments are required: image_path')
        if args.meofetch:
            meofetch(args)
        else:
            image_printer = run_cli()
            image_printer.print_text()
    except Exception as e:
        print(e)
        exit(1)

if __name__ == "__main__":
    ansifier_cli_main()

